;; Generated package description from gnorb.el  -*- no-byte-compile: t -*-
(define-package "gnorb" "1.6.5" "Glue code between Gnus, Org, and BBDB" '((cl-lib "0.5")) :url "http://elpa.gnu.org/packages/gnorb.html" :keywords '("mail" "org" "gnus" "bbdb" "todo" "task") :authors '(("Eric Abrahamsen" . "eric@ericabrahamsen.net")) :maintainer '("Eric Abrahamsen" . "eric@ericabrahamsen.net"))
